import casadi as ca
import matplotlib.pyplot as plt
import numpy as np

from rk4step import rk4step


# for people who cannot see an interactive plot, uncomment the following lines
# import matplotlib
# if matplotlib.get_backend() == 'agg':
#     matplotlib.use('WebAgg')
# print(f'backend: {matplotlib.get_backend()}')


## parameters required for the OCP
nx = ...  # state dimension
nu = ...  # control dimension
N = ...   # horizon
dt = ...  # discretization time step
N_rk4 = ...  # number of rk4 steps per discretization time step
x0_bar = ...   # initial state

## build the integrator
# ode
def dynamics(x, u):
    return ...

# discrete dynamics
x = ...
u = ...
x_next = ...

# integrator
F = ...


## NLP formulation
x_traj = ...  # vector of all states (1,..,N)
u_traj = ...  # vector of all controls (0,..,N-1)

x_traj_init = .1 * np.ones((nx*N, 1))  # intial guess
u_traj_init = .1 * np.ones((nu*N, 1))  # intial guess


# some utils
u_traj_per_stage = ca.vertsplit(u_traj, nu)
x_traj_per_stage = ca.vertsplit(x_traj, nx)
stage_cost = ca.Function('stage_cost', [x, u], [...])
terminal_cost = ca.Function('terminal_cost', [x], [...])

# build cost and constraints
x_current = x0_bar
constraints = []
cost = 0
for i in range(N):
    u_current = ...
    x_next = ...

    constraints.append(...)
    cost += stage_cost(x_current, u_current)

    x_current = ...
# terminal cost
cost += terminal_cost(x_current)
constraints = ca.vertcat(*constraints)


## lagrangian function
lam_traj = ...  # multiplier (1,..,N)
lagrangian = ...

# plot sparsity pattern of the Hessian of the lagrangian
lam_traj_per_stage = ca.vertsplit(lam_traj, nx)
z_traj = []
for i in range(N):
    # primal and dual variables in the required order 
    u_current = ...
    lam = ...
    x_next = ...
    z_traj.extend([u_current, lam, x_next])
z_traj = ca.vertcat(*z_traj)

(hess, grad) = ca.hessian(...)
hess_lagrangian = ca.Function('hess_lagrangian', [z_traj], [hess])
grad_lagrangian = ca.Function('grad_lagrangian', [z_traj], [grad])

plt.figure(1)
z_traj_mock = np.ones(z_traj.shape)
plt.spy(hess_lagrangian(z_traj_mock).full(), markersize=3)


## nlp solver
nlp_sim = {'x': ca.vertcat(x_traj, u_traj), 'f': cost, 'g':constraints}
solver_sim = ca.nlpsol('solver','ipopt', nlp_sim)

# solve nlp
sol_sim = solver_sim(
    x0=ca.vertcat(x_traj_init, u_traj_init),
    lbg=np.zeros(constraints.shape),
    ubg=np.zeros(constraints.shape),
)

## visualize solution
x_u_traj_opt_sim = sol_sim['x'].full().T
x_traj_opt_sim = ca.reshape(
    ca.vertcat(x0_bar, x_u_traj_opt_sim[:, 0:N*nx].T),
    (nx, N+1)
)
u_traj_opt_sim = x_u_traj_opt_sim[:, N*nx:None]

plt.figure(2)
plt.subplot(2,1,1)
plt.plot(...)
plt.title('state trajectory')
plt.legend([r'$\theta$', r'$\omega$'])

plt.subplot(2,1,2)
plt.step(...)
plt.title('control trajectory')
plt.legend([r'$\tau$'])
plt.xlabel('discrete time $k$')


plt.show()
